@@ -1,8 +1,8 @@
-/**
+/*
  *  Copyright 2004-2005 Paul Querna
  *  Copyright 2008, 2014 Nikos Mavrogiannopoulos
  *  Copyright 2011 Dash Shendy
- *  Copyright 2015 Thomas Klute
+ *  Copyright 2015-2016 Thomas Klute
  *
  *  Licensed under the Apache License, Version 2.0 (the "License");
  *  you may not use this file except in compliance with the License.
@@ -15,10 +15,11 @@
  *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  *  See the License for the specific language governing permissions and
  *  limitations under the License.
- *
  */
 
+#include "gnutls_config.h"
 #include "mod_gnutls.h"
+#include "gnutls_ocsp.h"
 #include "apr_lib.h"
 #include <gnutls/abstract.h>
 
@@ -92,335 +93,452 @@ const char static_dh_params[] = "-----BEGIN DH PARAMETERS-----\n"
         "Nd4jbVJfVHWbZeNy/NaO8g+nER+eSv9zAgEC\n"
         "-----END DH PARAMETERS-----\n";
 
-int mgs_load_files(apr_pool_t * p, server_rec * s)
+/*
+ * Clean up the various GnuTLS data structures allocated from
+ * mgs_load_files()
+ */
+static apr_status_t mgs_pool_free_credentials(void *arg)
+{
+    mgs_srvconf_rec *sc = (mgs_srvconf_rec *) arg;
+
+    if (sc->certs)
+    {
+        gnutls_certificate_free_credentials(sc->certs);
+        sc->certs = NULL;
+    }
+
+    if (sc->anon_creds)
+    {
+        gnutls_anon_free_server_credentials(sc->anon_creds);
+        sc->anon_creds = NULL;
+    }
+
+#ifdef ENABLE_SRP
+    if (sc->srp_creds)
+    {
+        gnutls_srp_free_server_credentials(sc->srp_creds);
+        sc->srp_creds = NULL;
+    }
+#endif
+
+    if (sc->dh_params)
+    {
+        gnutls_dh_params_deinit(sc->dh_params);
+        sc->dh_params = NULL;
+    }
+
+    for (unsigned int i = 0; i < sc->certs_x509_chain_num; i++)
+    {
+        gnutls_pcert_deinit(&sc->certs_x509_chain[i]);
+        gnutls_x509_crt_deinit(sc->certs_x509_crt_chain[i]);
+    }
+
+    if (sc->privkey_x509)
+    {
+        gnutls_privkey_deinit(sc->privkey_x509);
+        sc->privkey_x509 = NULL;
+    }
+
+    if (sc->ca_list)
+    {
+        for (unsigned int i = 0; i < sc->ca_list_size; i++)
+        {
+            gnutls_x509_crt_deinit(sc->ca_list[i]);
+        }
+        gnutls_free(sc->ca_list);
+        sc->ca_list = NULL;
+    }
+
+    if (sc->cert_pgp)
+    {
+        gnutls_pcert_deinit(&sc->cert_pgp[0]);
+        sc->cert_pgp = NULL;
+        gnutls_openpgp_crt_deinit(sc->cert_crt_pgp[0]);
+        sc->cert_crt_pgp = NULL;
+    }
+
+    if (sc->privkey_pgp)
+    {
+        gnutls_privkey_deinit(sc->privkey_pgp);
+        sc->privkey_pgp = NULL;
+#if GNUTLS_VERSION_NUMBER < 0x030312
+        gnutls_openpgp_privkey_deinit(sc->privkey_pgp_internal);
+        sc->privkey_pgp_internal = NULL;
+#endif
+    }
+
+    if (sc->pgp_list)
+    {
+        gnutls_openpgp_keyring_deinit(sc->pgp_list);
+        sc->pgp_list = NULL;
+    }
+
+    if (sc->priorities)
+    {
+        gnutls_priority_deinit(sc->priorities);
+        sc->priorities = NULL;
+    }
+
+    return APR_SUCCESS;
+}
+
+int mgs_load_files(apr_pool_t *pconf, apr_pool_t *ptemp, server_rec *s)
 {
     apr_pool_t *spool;
     const char *file;
     gnutls_datum_t data;
     int ret;
     mgs_srvconf_rec *sc =
-	(mgs_srvconf_rec *) ap_get_module_config(s->module_config,
-						 &gnutls_module);
+        (mgs_srvconf_rec *) ap_get_module_config(s->module_config,
+                                                 &gnutls_module);
 
-    apr_pool_create(&spool, p);
+    apr_pool_create(&spool, ptemp);
 
-    sc->cert_pgp = apr_pcalloc(p, sizeof(sc->cert_pgp[0]));
-    sc->cert_crt_pgp = apr_pcalloc(p, sizeof(sc->cert_crt_pgp[0]));
-    sc->certs_x509_chain =
-	apr_pcalloc(p, MAX_CHAIN_SIZE * sizeof(sc->certs_x509_chain[0]));
-    sc->certs_x509_crt_chain =
-	apr_pcalloc(p,
-		    MAX_CHAIN_SIZE * sizeof(sc->certs_x509_crt_chain[0]));
+    /* Cleanup function for the GnuTLS structures allocated below */
+    apr_pool_cleanup_register(pconf, sc, mgs_pool_free_credentials,
+                              apr_pool_cleanup_null);
 
-    ret = gnutls_certificate_allocate_credentials(&sc->certs);
-    if (ret < 0) {
-	ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-		     "GnuTLS: Failed to initialize" ": (%d) %s", ret,
-		     gnutls_strerror(ret));
-	ret = -1;
-	goto cleanup;
-    }
-
-    ret = gnutls_anon_allocate_server_credentials(&sc->anon_creds);
-    if (ret < 0) {
-	ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-		     "GnuTLS: Failed to initialize" ": (%d) %s", ret,
-		     gnutls_strerror(ret));
-	ret = -1;
-	goto cleanup;
-    }
-
-    /* Load SRP parameters */
-#ifdef ENABLE_SRP
-    ret = gnutls_srp_allocate_server_credentials(&sc->srp_creds);
-    if (ret < 0) {
-	ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-		     "GnuTLS: Failed to initialize" ": (%d) %s", ret,
-		     gnutls_strerror(ret));
-	ret = -1;
-	goto cleanup;
-    }
-
-    if (sc->srp_tpasswd_conf_file != NULL && sc->srp_tpasswd_file != NULL) {
-	ret = gnutls_srp_set_server_credentials_file
-	    (sc->srp_creds, sc->srp_tpasswd_file,
-	     sc->srp_tpasswd_conf_file);
-
-	if (ret < 0 && sc->enabled == GNUTLS_ENABLED_TRUE) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0,
-			 s,
-			 "GnuTLS: Host '%s:%d' is missing a "
-			 "SRP password or conf File!",
-			 s->server_hostname, s->port);
-	    ret = -1;
-	    goto cleanup;
-	}
-    }
-#endif
-
-    ret = gnutls_dh_params_init(&sc->dh_params);
-    if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to initialize"
-			 ": (%d) %s", ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-    }
-
-    /* Load DH parameters */
-    if (sc->dh_file) {
-	if (load_datum_from_file(spool, sc->dh_file, &data) != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Error Reading " "DH params '%s'", sc->dh_file);
-	    ret = -1;
-	    goto cleanup;
-	}
-
-	ret =
-	    gnutls_dh_params_import_pkcs3(sc->dh_params, &data,
-					  GNUTLS_X509_FMT_PEM);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Import "
-			 "DH params '%s': (%d) %s", sc->dh_file, ret,
-			 gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
-    } else {
-        gnutls_datum_t pdata = {
-            (void *) static_dh_params,
-            sizeof(static_dh_params)
-        };
-
-        ret = gnutls_dh_params_import_pkcs3(sc->dh_params, &pdata, GNUTLS_X509_FMT_PEM);
+    if (sc->certs == NULL)
+    {
+        ret = gnutls_certificate_allocate_credentials(&sc->certs);
         if (ret < 0) {
             ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-                    "GnuTLS: Unable to generate or load DH Params: (%d) %s",
-                    ret, gnutls_strerror(ret));
+                         "GnuTLS: Failed to initialize" ": (%d) %s", ret,
+                         gnutls_strerror(ret));
             ret = -1;
             goto cleanup;
         }
     }
 
-    if (sc->x509_cert_file != NULL) {
-	unsigned int chain_num, i;
-	unsigned format = GNUTLS_X509_FMT_PEM;
-
-	/* Load X.509 certificate */
-	if (strncmp(sc->x509_cert_file, "pkcs11:", 7) == 0) {
-	    gnutls_pkcs11_obj_t obj;
-
-	    file = sc->x509_cert_file;
-
-	    ret = gnutls_pkcs11_obj_init(&obj);
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Error Initializing PKCS #11 object");
-		ret = -1;
-		goto cleanup;
-	    }
-
-	    gnutls_pkcs11_obj_set_pin_function(obj, pin_callback, sc);
-
-	    ret = gnutls_pkcs11_obj_import_url(obj, file, GNUTLS_PKCS11_OBJ_FLAG_LOGIN);
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Error Importing PKCS #11 object: '%s': %s",
-			     file, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-
-	    format = GNUTLS_X509_FMT_DER;
-	    ret = gnutls_pkcs11_obj_export2(obj, &data);
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Error Exporting a PKCS #11 object: '%s': %s",
-			     file, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-
-	    gnutls_pkcs11_obj_deinit(obj);
-	} else {
-	    file = ap_server_root_relative(spool, sc->x509_cert_file);
-
-	    ret = gnutls_load_file(file, &data);
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Error Reading Certificate '%s': %s",
-			     file, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-	}
-
-	ret =
-	    gnutls_x509_crt_list_import2(&sc->certs_x509_crt_chain,
-					&chain_num, &data, format,
-					GNUTLS_X509_CRT_LIST_FAIL_IF_UNSORTED);
-	gnutls_free(data.data);
-	sc->certs_x509_chain_num = chain_num;
-
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Import Certificate Chain '%s': (%d) %s",
-			 file, ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
-
-	for (i = 0; i < chain_num; i++) {
-	    ret =
-		gnutls_pcert_import_x509(&sc->certs_x509_chain[i],
-					 sc->certs_x509_crt_chain[i], 0);
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Failed to Import pCertificate '%s': (%d) %s",
-			     file, ret, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-	}
-	sc->certs_x509_chain_num = chain_num;
+    if (sc->anon_creds == NULL)
+    {
+        ret = gnutls_anon_allocate_server_credentials(&sc->anon_creds);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize" ": (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
     }
 
-    if (sc->x509_key_file) {
-	ret = gnutls_privkey_init(&sc->privkey_x509);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to initialize: (%d) %s", ret,
-			 gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+    /* Load SRP parameters */
+#ifdef ENABLE_SRP
+    if (sc->srp_creds == NULL)
+    {
+        ret = gnutls_srp_allocate_server_credentials(&sc->srp_creds);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize" ": (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
+    }
 
-	if (gnutls_url_is_supported(sc->x509_key_file) != 0) {
-	    file = sc->x509_key_file;
+    if (sc->srp_tpasswd_conf_file != NULL && sc->srp_tpasswd_file != NULL)
+    {
+        ret = gnutls_srp_set_server_credentials_file
+            (sc->srp_creds, sc->srp_tpasswd_file,
+             sc->srp_tpasswd_conf_file);
 
-	    gnutls_privkey_set_pin_function(sc->privkey_x509, pin_callback,
-					    sc);
+        if (ret < 0 && sc->enabled == GNUTLS_ENABLED_TRUE) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Host '%s:%d' is missing a "
+                         "SRP password or conf File!",
+                         s->server_hostname, s->port);
+            ret = -1;
+            goto cleanup;
+        }
+    }
+#endif
 
-	    ret = gnutls_privkey_import_url(sc->privkey_x509, file, 0);
+    if (sc->dh_params == NULL)
+    {
+        ret = gnutls_dh_params_init(&sc->dh_params);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize"
+                         ": (%d) %s", ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Failed to Import Private Key URL '%s': (%d) %s",
-			     file, ret, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-	} else {
-	    file = ap_server_root_relative(spool, sc->x509_key_file);
+        /* Load DH parameters */
+        if (sc->dh_file)
+        {
+            if (load_datum_from_file(spool, sc->dh_file, &data) != 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Reading " "DH params '%s'", sc->dh_file);
+                ret = -1;
+                goto cleanup;
+            }
 
-	    if (load_datum_from_file(spool, file, &data) != 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Error Reading Private Key '%s'",
-			     file);
-		ret = -1;
-		goto cleanup;
-	    }
+            ret =
+                gnutls_dh_params_import_pkcs3(sc->dh_params, &data,
+                                              GNUTLS_X509_FMT_PEM);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Failed to Import "
+                             "DH params '%s': (%d) %s", sc->dh_file, ret,
+                             gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        } else {
+            gnutls_datum_t pdata = {
+                (void *) static_dh_params,
+                sizeof(static_dh_params)
+            };
 
-	    ret =
-		gnutls_privkey_import_x509_raw(sc->privkey_x509, &data,
-					       GNUTLS_X509_FMT_PEM, sc->pin,
-					       0);
+            ret = gnutls_dh_params_import_pkcs3(sc->dh_params, &pdata, GNUTLS_X509_FMT_PEM);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Unable to generate or load DH Params: (%d) %s",
+                             ret, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        }
+    }
 
-	    if (ret < 0) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Failed to Import Private Key '%s': (%d) %s",
-			     file, ret, gnutls_strerror(ret));
-		ret = -1;
-		goto cleanup;
-	    }
-	}
+    if (sc->x509_cert_file != NULL && sc->certs_x509_crt_chain == NULL)
+    {
+        sc->certs_x509_chain =
+            apr_pcalloc(pconf,
+                        MAX_CHAIN_SIZE * sizeof(sc->certs_x509_chain[0]));
+        sc->certs_x509_crt_chain =
+            apr_pcalloc(pconf,
+                        MAX_CHAIN_SIZE * sizeof(sc->certs_x509_crt_chain[0]));
+        unsigned int chain_num = MAX_CHAIN_SIZE;
+        unsigned format = GNUTLS_X509_FMT_PEM;
+
+        /* Load X.509 certificate */
+        if (strncmp(sc->x509_cert_file, "pkcs11:", 7) == 0) {
+            gnutls_pkcs11_obj_t obj;
+
+            file = sc->x509_cert_file;
+
+            ret = gnutls_pkcs11_obj_init(&obj);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Initializing PKCS #11 object");
+                ret = -1;
+                goto cleanup;
+            }
+
+            gnutls_pkcs11_obj_set_pin_function(obj, pin_callback, sc);
+
+            ret = gnutls_pkcs11_obj_import_url(obj, file,
+                                               GNUTLS_PKCS11_OBJ_FLAG_LOGIN);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Importing PKCS #11 object: "
+                             "'%s': %s",
+                             file, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+
+            format = GNUTLS_X509_FMT_DER;
+            ret = gnutls_pkcs11_obj_export2(obj, &data);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Exporting a PKCS #11 object: "
+                             "'%s': %s",
+                             file, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+
+            gnutls_pkcs11_obj_deinit(obj);
+        } else {
+            file = ap_server_root_relative(spool, sc->x509_cert_file);
+
+            ret = gnutls_load_file(file, &data);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Reading Certificate '%s': %s",
+                             file, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        }
+
+        ret = gnutls_x509_crt_list_import(sc->certs_x509_crt_chain,
+                                          &chain_num, &data, format,
+                                          GNUTLS_X509_CRT_LIST_FAIL_IF_UNSORTED);
+        gnutls_free(data.data);
+        sc->certs_x509_chain_num = chain_num;
+
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to Import Certificate Chain "
+                         "'%s': (%d) %s",
+                         file, ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
+
+        for (unsigned int i = 0; i < chain_num; i++)
+        {
+            ret =
+                gnutls_pcert_import_x509(&sc->certs_x509_chain[i],
+                                         sc->certs_x509_crt_chain[i], 0);
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Failed to Import pCertificate "
+                             "'%s': (%d) %s",
+                             file, ret, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        }
+        sc->certs_x509_chain_num = chain_num;
+    }
+
+    if (sc->x509_key_file && sc->privkey_x509 == NULL)
+    {
+        ret = gnutls_privkey_init(&sc->privkey_x509);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize: (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
+
+        if (gnutls_url_is_supported(sc->x509_key_file) != 0) {
+            file = sc->x509_key_file;
+
+            gnutls_privkey_set_pin_function(sc->privkey_x509, pin_callback,
+                                            sc);
+
+            ret = gnutls_privkey_import_url(sc->privkey_x509, file, 0);
+
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Failed to Import Private Key URL "
+                             "'%s': (%d) %s",
+                             file, ret, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        } else {
+            file = ap_server_root_relative(spool, sc->x509_key_file);
+
+            if (load_datum_from_file(spool, file, &data) != 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Error Reading Private Key '%s'",
+                             file);
+                ret = -1;
+                goto cleanup;
+            }
+
+            ret =
+                gnutls_privkey_import_x509_raw(sc->privkey_x509, &data,
+                                               GNUTLS_X509_FMT_PEM, sc->pin,
+                                               0);
+
+            if (ret < 0) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Failed to Import Private Key "
+                             "'%s': (%d) %s",
+                             file, ret, gnutls_strerror(ret));
+                ret = -1;
+                goto cleanup;
+            }
+        }
     }
 
     /* Load the X.509 CA file */
-    if (sc->x509_ca_file) {
-	if (load_datum_from_file(spool, sc->x509_ca_file, &data) != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Error Reading " "Client CA File '%s'",
-			 sc->x509_ca_file);
-	    ret = -1;
-	    goto cleanup;
-	}
+    if (sc->x509_ca_file)
+    {
+        if (load_datum_from_file(spool, sc->x509_ca_file, &data) != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Error Reading " "Client CA File '%s'",
+                         sc->x509_ca_file);
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret = gnutls_x509_crt_list_import2(&sc->ca_list, &sc->ca_list_size,
-					 &data, GNUTLS_X509_FMT_PEM, 0);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to load "
-			 "Client CA File '%s': (%d) %s", sc->x509_ca_file,
-			 ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_x509_crt_list_import2(&sc->ca_list, &sc->ca_list_size,
+                                           &data, GNUTLS_X509_FMT_PEM, 0);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to load "
+                         "Client CA File '%s': (%d) %s", sc->x509_ca_file,
+                         ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
     }
 
-    if (sc->pgp_cert_file) {
-	if (load_datum_from_file(spool, sc->pgp_cert_file, &data) != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Error Reading " "Certificate '%s'",
-			 sc->pgp_cert_file);
-	    ret = -1;
-	    goto cleanup;
-	}
+    if (sc->pgp_cert_file && sc->cert_pgp == NULL)
+    {
+        sc->cert_pgp = apr_pcalloc(pconf, sizeof(sc->cert_pgp[0]));
+        sc->cert_crt_pgp = apr_pcalloc(pconf, sizeof(sc->cert_crt_pgp[0]));
 
-	ret = gnutls_openpgp_crt_init(&sc->cert_crt_pgp[0]);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Init "
-			 "PGP Certificate: (%d) %s", ret,
-			 gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        if (load_datum_from_file(spool, sc->pgp_cert_file, &data) != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Error Reading " "Certificate '%s'",
+                         sc->pgp_cert_file);
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret =
-	    gnutls_openpgp_crt_import(sc->cert_crt_pgp[0], &data,
-				      GNUTLS_OPENPGP_FMT_BASE64);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Import "
-			 "PGP Certificate: (%d) %s", ret,
-			 gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_openpgp_crt_init(&sc->cert_crt_pgp[0]);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to Init "
+                         "PGP Certificate: (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret =
-	    gnutls_pcert_import_openpgp(sc->cert_pgp, sc->cert_crt_pgp[0],
-					0);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Import "
-			 "PGP pCertificate: (%d) %s", ret,
-			 gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_openpgp_crt_import(sc->cert_crt_pgp[0], &data,
+                                        GNUTLS_OPENPGP_FMT_BASE64);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to Import "
+                         "PGP Certificate: (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
+
+        ret = gnutls_pcert_import_openpgp(sc->cert_pgp,
+                                          sc->cert_crt_pgp[0], 0);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to Import "
+                         "PGP pCertificate: (%d) %s", ret,
+                         gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
     }
 
     /* Load the PGP key file */
-    if (sc->pgp_key_file) {
-	if (load_datum_from_file(spool, sc->pgp_key_file, &data) != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Error Reading " "Private Key '%s'",
-			 sc->pgp_key_file);
-	    ret = -1;
-	    goto cleanup;
-	}
+    if (sc->pgp_key_file && sc->privkey_pgp == NULL)
+    {
+        if (load_datum_from_file(spool, sc->pgp_key_file, &data) != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Error Reading " "Private Key '%s'",
+                         sc->pgp_key_file);
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret = gnutls_privkey_init(&sc->privkey_pgp);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to initialize"
-			 ": (%d) %s", ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_privkey_init(&sc->privkey_pgp);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize"
+                         ": (%d) %s", ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
 #if GNUTLS_VERSION_NUMBER < 0x030312
         /* GnuTLS versions before 3.3.12 contain a bug in
@@ -432,25 +550,25 @@ int mgs_load_files(apr_pool_t * p, server_rec * s)
          * gnutls_openpgp_privkey_import and
          * gnutls_privkey_import_openpgp. */
         ret = gnutls_openpgp_privkey_init(&sc->privkey_pgp_internal);
-	if (ret != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to initialize "
-			 "PGP Private Key '%s': (%d) %s",
-			 sc->pgp_key_file, ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        if (ret != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize "
+                         "PGP Private Key '%s': (%d) %s",
+                         sc->pgp_key_file, ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
         ret = gnutls_openpgp_privkey_import(sc->privkey_pgp_internal, &data,
                                             GNUTLS_OPENPGP_FMT_BASE64, NULL, 0);
-	if (ret != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to Import "
-			 "PGP Private Key '%s': (%d) %s",
-			 sc->pgp_key_file, ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        if (ret != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to Import "
+                         "PGP Private Key '%s': (%d) %s",
+                         sc->pgp_key_file, ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
         ret = gnutls_privkey_import_openpgp(sc->privkey_pgp,
                                             sc->privkey_pgp_internal, 0);
@@ -460,77 +578,79 @@ int mgs_load_files(apr_pool_t * p, server_rec * s)
                          "GnuTLS: Failed to assign PGP Private Key '%s' "
                          "to gnutls_privkey_t structure: (%d) %s",
                          sc->pgp_key_file, ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+            ret = -1;
+            goto cleanup;
+        }
 #else
         ret = gnutls_privkey_import_openpgp_raw(sc->privkey_pgp, &data,
                                                 GNUTLS_OPENPGP_FMT_BASE64,
                                                 NULL, NULL);
-	if (ret != 0)
+        if (ret != 0)
         {
             ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
                          "GnuTLS: Failed to Import "
                          "PGP Private Key '%s': (%d) %s",
                          sc->pgp_key_file, ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+            ret = -1;
+            goto cleanup;
+        }
 #endif
     }
 
     /* Load the keyring file */
-    if (sc->pgp_ring_file) {
-	if (load_datum_from_file(spool, sc->pgp_ring_file, &data) != 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Error Reading " "Keyring File '%s'",
-			 sc->pgp_ring_file);
-	    ret = -1;
-	    goto cleanup;
-	}
+    if (sc->pgp_ring_file && sc->pgp_list == NULL)
+    {
+        if (load_datum_from_file(spool, sc->pgp_ring_file, &data) != 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Error Reading " "Keyring File '%s'",
+                         sc->pgp_ring_file);
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret = gnutls_openpgp_keyring_init(&sc->pgp_list);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to initialize"
-			 "keyring: (%d) %s", ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_openpgp_keyring_init(&sc->pgp_list);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to initialize"
+                         "keyring: (%d) %s", ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
 
-	ret = gnutls_openpgp_keyring_import(sc->pgp_list, &data,
-					   GNUTLS_OPENPGP_FMT_BASE64);
-	if (ret < 0) {
-	    ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			 "GnuTLS: Failed to load "
-			 "Keyring File '%s': (%d) %s", sc->pgp_ring_file,
-			 ret, gnutls_strerror(ret));
-	    ret = -1;
-	    goto cleanup;
-	}
+        ret = gnutls_openpgp_keyring_import(sc->pgp_list, &data,
+                                            GNUTLS_OPENPGP_FMT_BASE64);
+        if (ret < 0) {
+            ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                         "GnuTLS: Failed to load "
+                         "Keyring File '%s': (%d) %s", sc->pgp_ring_file,
+                         ret, gnutls_strerror(ret));
+            ret = -1;
+            goto cleanup;
+        }
     }
 
-    if (sc->priorities_str) {
+    if (sc->priorities_str && sc->priorities == NULL)
+    {
         const char *err;
         ret = gnutls_priority_init(&sc->priorities, sc->priorities_str, &err);
 
-	if (ret < 0) {
-	    if (ret == GNUTLS_E_INVALID_REQUEST) {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: Syntax error parsing priorities string at: %s",
-			     err);
-	    } else {
-		ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
-			     "GnuTLS: error parsing priorities string");
+        if (ret < 0) {
+            if (ret == GNUTLS_E_INVALID_REQUEST) {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: Syntax error parsing priorities string at: %s",
+                             err);
+            } else {
+                ap_log_error(APLOG_MARK, APLOG_STARTUP, 0, s,
+                             "GnuTLS: error parsing priorities string");
 
-	    }
-	    ret = -1;
-	    goto cleanup;
-	}
+            }
+            ret = -1;
+            goto cleanup;
+        }
     }
 
     ret = 0;
-  cleanup:
+ cleanup:
     apr_pool_destroy(spool);
 
     return ret;
@@ -726,28 +846,40 @@ const char *mgs_set_cache(cmd_parms * parms, void *dummy __attribute__((unused))
     return NULL;
 }
 
-const char *mgs_set_cache_timeout(cmd_parms * parms, void *dummy __attribute__((unused)),
-        const char *arg) {
-    int argint;
-    const char *err;
-    mgs_srvconf_rec *sc =
-	(mgs_srvconf_rec *) ap_get_module_config(parms->server->
-						 module_config,
-						 &gnutls_module);
+const char *mgs_set_timeout(cmd_parms * parms,
+                            void *dummy __attribute__((unused)),
+                            const char *arg)
+{
+    apr_int64_t argint = apr_atoi64(arg);
+    /* timeouts cannot be negative */
+    if (argint < 0)
+        return apr_psprintf(parms->pool, "%s: Invalid argument",
+                            parms->directive->directive);
 
-    if ((err = ap_check_cmd_context(parms, GLOBAL_ONLY))) {
-	return err;
-    }
+    mgs_srvconf_rec *sc = (mgs_srvconf_rec *)
+        ap_get_module_config(parms->server->module_config, &gnutls_module);
 
-    argint = atoi(arg);
-
-    if (argint < 0) {
-	return "GnuTLSCacheTimeout: Invalid argument";
-    } else if (argint == 0) {
-	sc->cache_timeout = 0;
-    } else {
-	sc->cache_timeout = apr_time_from_sec(argint);
+    if (!apr_strnatcasecmp(parms->directive->directive, "GnuTLSCacheTimeout"))
+    {
+        const char *err;
+        if ((err = ap_check_cmd_context(parms, GLOBAL_ONLY)))
+            return err;
+        sc->cache_timeout = apr_time_from_sec(argint);
     }
+    else if (!apr_strnatcasecmp(parms->directive->directive,
+                                "GnuTLSOCSPCacheTimeout"))
+        sc->ocsp_cache_time = apr_time_from_sec(argint);
+    else if (!apr_strnatcasecmp(parms->directive->directive,
+                                "GnuTLSOCSPFailureTimeout"))
+        sc->ocsp_failure_timeout = apr_time_from_sec(argint);
+    else if (!apr_strnatcasecmp(parms->directive->directive,
+                                "GnuTLSOCSPSocketTimeout"))
+        sc->ocsp_socket_timeout = apr_time_from_sec(argint);
+    else
+        /* Can't happen unless there's a serious bug in mod_gnutls or Apache */
+        return apr_psprintf(parms->pool,
+                            "mod_gnutls: %s called for invalid option '%s'",
+                            __func__, parms->directive->directive);
 
     return NULL;
 }
@@ -949,17 +1081,35 @@ static mgs_srvconf_rec *_mgs_config_server_create(apr_pool_t * p,
     sc->enabled = GNUTLS_ENABLED_UNSET;
 
     sc->privkey_x509 = NULL;
-    sc->privkey_pgp = NULL;
+    sc->anon_creds = NULL;
+#ifdef ENABLE_SRP
+    sc->srp_creds = NULL;
+#endif
+    sc->certs = NULL;
+    sc->certs_x509_chain = NULL;
+    sc->certs_x509_crt_chain = NULL;
     sc->certs_x509_chain_num = 0;
     sc->p11_modules = NULL;
     sc->pin = NULL;
+
+    sc->cert_pgp = NULL;
+    sc->cert_crt_pgp = NULL;
+    sc->privkey_pgp = NULL;
+#if GNUTLS_VERSION_NUMBER < 0x030312
+    sc->privkey_pgp_internal = NULL;
+#endif
+    sc->pgp_list = NULL;
+
     sc->priorities_str = NULL;
-    sc->cache_timeout = -1;	/* -1 means "unset" */
+    sc->cache_timeout = MGS_TIMEOUT_UNSET;
     sc->cache_type = mgs_cache_unset;
     sc->cache_config = NULL;
+    sc->cache = NULL;
     sc->tickets = GNUTLS_ENABLED_UNSET;
     sc->priorities = NULL;
     sc->dh_params = NULL;
+    sc->ca_list = NULL;
+    sc->ca_list_size = 0;
     sc->proxy_enabled = GNUTLS_ENABLED_UNSET;
     sc->export_certificates_size = -1;
     sc->client_verify_method = mgs_cvm_unset;
@@ -969,7 +1119,18 @@ static mgs_srvconf_rec *_mgs_config_server_create(apr_pool_t * p,
     sc->proxy_x509_ca_file = NULL;
     sc->proxy_x509_crl_file = NULL;
     sc->proxy_priorities_str = NULL;
+    sc->proxy_x509_creds = NULL;
+    sc->anon_client_creds = NULL;
     sc->proxy_priorities = NULL;
+    sc->proxy_x509_tl = NULL;
+
+    sc->ocsp_staple = GNUTLS_ENABLED_UNSET;
+    sc->ocsp_check_nonce = GNUTLS_ENABLED_UNSET;
+    sc->ocsp_response_file = NULL;
+    sc->ocsp_mutex = NULL;
+    sc->ocsp_cache_time = MGS_TIMEOUT_UNSET;
+    sc->ocsp_failure_timeout = MGS_TIMEOUT_UNSET;
+    sc->ocsp_socket_timeout = MGS_TIMEOUT_UNSET;
 
 /* this relies on GnuTLS never changing the gnutls_certificate_request_t enum to define -1 */
     sc->client_verify_mode = -1;
@@ -1027,12 +1188,13 @@ void *mgs_config_server_merge(apr_pool_t * p, void *BASE, void *ADD)
     gnutls_srvconf_merge(proxy_priorities_str, NULL);
     gnutls_srvconf_merge(proxy_priorities, NULL);
 
-    /* FIXME: the following items are pre-allocated, and should be
-     * properly disposed of before assigning in order to avoid leaks;
-     * so at the moment, we can't actually have them in the config.
-     * what happens during de-allocation? */
-    /* TODO: mgs_load_files takes care of most of these now, verify
-     * and clean up the following lines */
+    gnutls_srvconf_merge(ocsp_staple, GNUTLS_ENABLED_UNSET);
+    gnutls_srvconf_merge(ocsp_check_nonce, GNUTLS_ENABLED_UNSET);
+    gnutls_srvconf_assign(ocsp_response_file);
+    gnutls_srvconf_merge(ocsp_cache_time, MGS_TIMEOUT_UNSET);
+    gnutls_srvconf_merge(ocsp_failure_timeout, MGS_TIMEOUT_UNSET);
+    gnutls_srvconf_merge(ocsp_socket_timeout, MGS_TIMEOUT_UNSET);
+
     gnutls_srvconf_assign(ca_list);
     gnutls_srvconf_assign(ca_list_size);
     gnutls_srvconf_assign(cert_pgp);
